'use strict';

var _createClass = require('babel-runtime/helpers/create-class')['default'];

var _classCallCheck = require('babel-runtime/helpers/class-call-check')['default'];

var _interopRequireDefault = require('babel-runtime/helpers/interop-require-default')['default'];

Object.defineProperty(exports, '__esModule', {
    value: true
});

var _baobab = require('baobab');

var _baobab2 = _interopRequireDefault(_baobab);

var _lodashIsequal = require('lodash.isequal');

var _lodashIsequal2 = _interopRequireDefault(_lodashIsequal);

var State = (function () {
    function State(initialState, options) {
        if (initialState === undefined) initialState = {};

        _classCallCheck(this, State);

        this._tree = new _baobab2['default'](initialState, options);
        this._watchingQueue = [];
        this._watchingPaths = [];
        this._fetchers = [];
        this._senders = [];

        this._onPathGet = this._onPathGet.bind(this);
        this._onPathWrite = this._onPathWrite.bind(this);
    }

    /**
     * Fetchers API
     */

    _createClass(State, [{
        key: '_onPathGet',
        value: function _onPathGet(e) {
            var path = e.data.path;

            this._fetchers.forEach(this._processFetcherWithPath.bind(this, path));
        }
    }, {
        key: '_isPathMatchedBy',
        value: function _isPathMatchedBy(cursorPath, matchersFactories) {
            return matchersFactories.some(function (matchersFactory) {
                return matchersFactory(cursorPath).some(function (matcher) {
                    return matcher.path.every(function (chunk, i) {
                        return (0, _lodashIsequal2['default'])(chunk, cursorPath[i]);
                    });
                });
            });
        }
    }, {
        key: '_isPathMatched',
        value: function _isPathMatched(cursorPath) {
            return this._fetchers.some(this._isPathMatchedBy.bind(this, cursorPath));
        }
    }, {
        key: '_processFetcherWithPath',
        value: function _processFetcherWithPath(cursorPath, matchersFactories) {
            var isDataExists = this.exists(cursorPath);

            matchersFactories.forEach(function (matchersFactory) {
                matchersFactory(cursorPath).forEach(function (matcher) {
                    var matched = matcher.path.every(function (chunk, i) {
                        return chunk === cursorPath[i];
                    });

                    // if cursor path is matched by some fetcher and
                    // such data is not exists yet
                    if (matched && !isDataExists) {
                        matcher.callback();
                    }
                });
            });
        }
    }, {
        key: '_registerFetcher',
        value: function _registerFetcher(fetcher) {
            var _this = this;

            // add `get`-listener if there is no fetchers yet
            if (this._fetchers.length === 0) {
                this._tree.on('get', this._onPathGet);
            }

            // move matched by new fetcher paths from queue
            this._watchingQueue = this._watchingQueue.filter(function (queuePath) {
                if (_this._isPathMatchedBy(queuePath, fetcher)) {
                    _this._watchingPaths.push(queuePath);

                    return false;
                }

                return true;
            });

            // process all watching paths with the new fetcher
            this._watchingPaths.forEach(function (watchingPath) {
                _this._processFetcherWithPath(watchingPath, fetcher);
            });

            // store fetcher
            this._fetchers.push(fetcher);
        }
    }, {
        key: '_unregisterFetcher',
        value: function _unregisterFetcher(fetcher) {
            var _this2 = this;

            // remove fetcher from registered fetchers
            this._fetchers = this._fetchers.filter(function (registeredFetcher) {
                return registeredFetcher !== fetcher;
            });

            // move watched by nobody paths to queue
            this._watchingPaths = this._watchingPaths.filter(function (watchingPath) {
                if (!_this2._isPathMatched(watchingPath)) {
                    _this2._watchingQueue.push(watchingPath);

                    return false;
                }

                return true;
            });

            // remove `get`-listener if there is no fetchers anymore
            if (this._fetchers.length === 0) {
                this._tree.off('get', this._onPathGet);
            }
        }
    }, {
        key: '_addToWatchingQueue',
        value: function _addToWatchingQueue(cursorPath) {
            // only if path is not matched by registered fetchers
            if (this._isPathMatched(cursorPath)) {
                this._watchingPaths.push(cursorPath);
            } else {
                this._watchingQueue.push(cursorPath);
            }
        }
    }, {
        key: '_removeFromWatchingQueue',
        value: function _removeFromWatchingQueue(cursorPath) {
            // remove from watching paths if path is matched by any registered fetchers
            if (this._isPathMatched(cursorPath)) {
                this._watchingPaths.splice(this._watchingPaths.indexOf(cursorPath), 1);

                // otherwise remove from queue
            } else {
                    this._watchingQueue.splice(this._watchingQueue.indexOf(cursorPath), 1);
                }
        }

        /**
         * Senders API
         */
    }, {
        key: '_onPathWrite',
        value: function _onPathWrite(e) {
            var path = e.data.path;

            this._senders.forEach(this._processSenderWithPath.bind(this, path));
        }
    }, {
        key: '_processSenderWithPath',
        value: function _processSenderWithPath(cursorPath, matchersFactories) {
            var isDataExists = this.exists(cursorPath);

            // it's intentionally undefined by default
            // because Baobab returns undefined if there is no value in path
            var value = undefined;

            // we retrieve value only if it is exists
            // so DataFetcher wouldn't trigger when we don't want to
            if (isDataExists) {
                value = this.getIn(cursorPath);
            }

            matchersFactories.forEach(function (matchersFactory) {
                matchersFactory(cursorPath).forEach(function (matcher) {
                    var matched = matcher.path.every(function (chunk, i) {
                        return chunk === cursorPath[i];
                    });

                    // trigger callback if path is matched
                    if (matched) {
                        matcher.callback(value);
                    }
                });
            });
        }
    }, {
        key: '_registerSender',
        value: function _registerSender(sender) {
            // add `write`-listener if there is no senders yet
            if (this._senders.length === 0) {
                this._tree.on('write', this._onPathWrite);
            }

            // store sender
            this._senders.push(sender);
        }
    }, {
        key: '_unregisterSender',
        value: function _unregisterSender(sender) {
            // remove sender from registered senders
            this._senders = this._senders.filter(function (registeredSender) {
                return registeredSender !== sender;
            });

            // remove `write`-listener if there is no senders anymore
            if (this._senders.length === 0) {
                this._tree.off('write', this._onPathWrite);
            }
        }

        /**
         * Public API
         */
    }, {
        key: 'getTree',
        value: function getTree() {
            return this._tree;
        }
    }, {
        key: 'get',
        value: function get() {
            return this._tree.get();
        }
    }, {
        key: 'getIn',
        value: function getIn(cursorPath) {
            return this._tree.get(cursorPath);
        }
    }, {
        key: 'set',
        value: function set(data) {
            this._tree.set(data);
        }
    }, {
        key: 'setIn',
        value: function setIn(cursorPath, data) {
            this._tree.set(cursorPath, data);
        }
    }, {
        key: 'unset',
        value: function unset() {
            this._tree.unset();
        }
    }, {
        key: 'unsetIn',
        value: function unsetIn(cursorPath) {
            this._tree.unset(cursorPath);
        }
    }, {
        key: 'exists',
        value: function exists(path) {
            return this._tree.select(path).exists();
        }
    }]);

    return State;
})();

exports['default'] = State;
module.exports = exports['default'];